# http://www.earlevel.com/main/2012/11/26/biquad-c-source-code/

import jax.numpy as jnp
from jax import jit, lax

from jaxdsp.param import Param

NAME = "BiQuad Lowpass Filter"
PARAMS = [Param("resonance", 0.7), Param("cutoff", 0.49, 0.0, 0.49)]
PRESETS = {}


def init_state(buffer_size=20):
    return {
        "a0": 1.0,
        "a1": 0.0,
        "a2": 0.0,
        "b1": 0.0,
        "b2": 0.0,
        "z1": 0.0,
        "z2": 0.0,
    }


@jit
def tick(carry, x):
    _, state = carry
    out = x * state["a0"] + state["z1"]
    state["z1"] = x * state["a1"] + state["z2"] - state["b1"] * out
    state["z2"] = x * state["a2"] - state["b2"] * out

    return carry, out


@jit
def tick_buffer(carry, X):
    params, state = carry
    K = jnp.tan(jnp.pi * params["cutoff"])
    norm = 1.0 / (1.0 + K / params["resonance"] + K * K)
    state["a0"] = K * K * norm
    state["a1"] = 2.0 * state["a0"]
    state["a2"] = state["a0"]
    state["b1"] = 2.0 * (K * K - 1) * norm
    state["b2"] = (1.0 - K / params["resonance"] + K * K) * norm
    return lax.scan(tick, carry, X)



import jax.numpy as jnp
from jax import jit
import jax_spectral

# Port of https://github.com/magenta/ddsp/blob/master/ddsp/spectral_ops.py#L33,
# but using https://github.com/cifkao/jax-spectral instead of tf.signal.stft
# Note: keep an eye out for jax to implement this in
# [jax.scipy.signal](https://jax.readthedocs.io/en/latest/jax.scipy.html#module-jax.scipy.signal)
def stft(audio, sample_rate=16000, frame_size=2048, overlap=0.75, pad_end=True):
    assert frame_size * overlap % 2.0 == 0.0

    return jax_spectral.spectral.stft(
        audio,
        # sample_rate, # TODO need this
        nperseg=int(frame_size),
        noverlap=int(overlap),
        # nfft=int(frame_size),
        padded=pad_end,
    )


# Port of https://github.com/magenta/ddsp/blob/master/ddsp/spectral_ops.py#L76
def magnitute_spectrogram(audio, size=2048, overlap=0.75, pad_end=True):
    return jnp.abs(stft(audio, frame_size=size, overlap=overlap, pad_end=pad_end)[2])


def safe_log(X, eps=1e-5):
    """Avoid taking the log of a non-positive number."""
    return jnp.log(jnp.where(X <= eps, eps, X))


def norm(X):
    return jnp.sqrt((X ** 2).sum())


def distance(X, Y, kind="L1"):
    kind = kind.upper()
    if kind == "L1":
        return jnp.mean(jnp.abs(Y - X))
    elif kind == "L2":
        return jnp.mean((Y - X) ** 2)
    elif kind == "COSINE":
        return (Y @ X.T) / (norm(Y) * norm(X))
    else:
        raise ValueError(f"Distance type ({kind}), must be 'L1, 'L2', or 'COSINE'")


# Spectral losses based on https://github.com/magenta/ddsp/blob/master/ddsp/losses.py#L132
# Doesn't support `loudness` from original ddsp.
# ddsp also has lots more experimental losses (that are worth investigating at some point!)
loss_fn_for_label = {
    # Compare raw samples (time domain).
    "sample": lambda X, Y, distance_type: distance(X, Y, distance_type),
    # Compare linear magnitudes of spectrograms.
    # Core audio similarity loss. More sensitive to peak magnitudes than log magnitudes.
    "magnitude": lambda X_mag, Y_mag, distance_type: distance(
        X_mag, Y_mag, distance_type
    ),
    # Compare log magnitudes of spectrograms.
    # Core audio similarity loss. More sensitive to quiet magnitudes than linear magnitudes.
    "log_magnitude": lambda X_mag, Y_mag, distance_type: distance(
        safe_log(X_mag), safe_log(Y_mag), distance_type
    ),
    # Compare the first finite difference of spectrograms in time.
    # Emphasizes changes of magnitude in time, such as at transients.
    "delta_time": lambda X_mag, Y_mag, distance_type: distance(
        jnp.diff(X_mag, axis=1), jnp.diff(Y_mag, axis=1), distance_type
    ),
    # Compare the first finite difference of spectrograms in frequency.
    # Emphasizes changes of magnitude in frequency, such as at the boundaries of a stack of harmonics.
    "delta_freq": lambda X_mag, Y_mag, distance_type: distance(
        jnp.diff(X_mag, axis=2), jnp.diff(Y_mag, axis=2), distance_type
    ),
    # Compare the cumulative sum of spectrograms across frequency for each slice in time.
    # Similar to a 1-D Wasserstein loss.
    # This hopefully provides a non-vanishing gradient to push two non-overlapping sinusoids towards each other.
    "cumsum_freq": lambda X_mag, Y_mag, distance_type: distance(
        jnp.cumsum(X_mag, axis=2), jnp.cumsum(Y_mag, axis=2), distance_type
    ),
}

sample_loss_labels = ["sample"]
frequency_loss_labels = [
    "magnitude",
    "log_magnitude",
    "delta_time",
    "delta_freq",
    "cumsum_freq",
]


class LossOptions:
    def __init__(
        self,
        weights={
            "sample": 1.0,
            "magnitude": 0.0,
            "log_magnitude": 0.0,
            "delta_time": 0.0,
            "delta_freq": 0.0,
            "cumsum_freq": 0.0,
        },
        distance_types={
            "sample": "L1",
            "frequency": "L1",
        },
        # Note: removing smaller fft sizes seems to get rid of some small non-convex "bumps"
        # in the loss curve for a sine wave with a target frequency param.
        # Also, removing 2048 size since it's just so darn slow.
        # fft_sizes=(2048, 1024, 512, 256, 128, 64),
        fft_sizes=(1024, 512, 256),
    ):
        """Args:
        weights: Dict of loss labels to relative weighting of that loss.
            (See `loss_fn_for_label` above for details on loss types.)
        fft_sizes: Compare spectrograms at each of this list of fft sizes.
            Each spectrogram has a time-frequency resolution trade-off based on fft size,
            so comparing multiple scales allows multiple resolutions.
        distance_types: Dict of "sample" and/or "frequency", with possible values: 'L1', 'L2', or 'COSINE'
        """
        self.weights_options = weights or {}
        distance_types = distance_types or {}
        self.fft_sizes = fft_sizes or (1024, 512, 256)
        self.sample_distance_type = distance_types.get("sample") or "L1"
        self.frequency_distance_type = distance_types.get("frequency") or "L1"
        self.sample_weights = [
            (label, weight)
            for label, weight in self.weights_options.items()
            if weight and weight > 0 and label in sample_loss_labels
        ]
        self.frequency_weights = [
            (label, weight)
            for label, weight in self.weights_options.items()
            if weight and weight > 0 and label in frequency_loss_labels
        ]

    def serialize(self):
        return {
            # Every key should be present in serialized options.
            "weights": {
                label: self.weights_options.get(label) or 0.0
                for label in sample_loss_labels
            }
            | {
                label: self.weights_options.get(label) or 0.0
                for label in frequency_loss_labels
            },
            "distance_types": {
                "sample": self.sample_distance_type,
                "frequency": self.frequency_distance_type,
            },
            "fft_sizes": list(self.fft_sizes),
        }


def loss_fn(X, Y, opts):
    loss = sum(
        weight * loss_fn_for_label[label](X, Y, opts.sample_distance_type)
        for label, weight in opts.sample_weights
    )
    if len(opts.frequency_weights) > 0:
        for fft_size in opts.fft_sizes:
            # TODO janky. Should have the same dimension reqs for time & freq
            if len(X.shape) == 1:
                X = jnp.expand_dims(X, 0)
            if len(Y.shape) == 1:
                Y = jnp.expand_dims(Y, 0)
            X_mag = magnitute_spectrogram(X, size=fft_size)
            Y_mag = magnitute_spectrogram(Y, size=fft_size)
            loss += sum(
                weight
                * loss_fn_for_label[label](X_mag, Y_mag, opts.frequency_distance_type)
                for label, weight in opts.frequency_weights
            )
    return loss


def correlation(X, Y):
    """Based on https://github.com/numpy/numpy/issues/2310#issuecomment-314265666"""
    c = jnp.correlate(
        (X - jnp.mean(X)) / (jnp.std(X) * len(X)), (Y - jnp.mean(Y)) / (jnp.std(Y))
    )[0]
    return 1 - (1 + c) / 2


# TODO error-to-signal ratio (ESR)
#   MSE normalized by signal energy: "You need to be more accurate when your signal is very quiet"



from jaxdsp.processors import (
    allpass_filter,
    clip,
    delay_line,
    fir_filter,
    freeverb,
    iir_filter,
    biquad_lowpass,
    lowpass_feedback_comb_filter,
    sine_wave,
)

all_processors = [
    allpass_filter,
    clip,
    delay_line,
    fir_filter,
    freeverb,
    iir_filter,
    biquad_lowpass,
    lowpass_feedback_comb_filter,
    sine_wave,
]
processor_by_name = {processor.NAME: processor for processor in all_processors}


def default_param_values(processor):
    return (
        {param.name: param.default_value for param in processor.PARAMS}
        if processor
        else None
    )


def serialize_processor(processor, params=None):
    if not processor:
        return None

    return {
        "name": processor.NAME,
        "param_definitions": [param.serialize() for param in processor.PARAMS],
        "presets": processor.PRESETS,
        "params": params or default_param_values(processor),
    }


def params_to_unit_scale(params, processor_name):
    if not params or not processor_name:
        return None

    if isinstance(processor_name, list):
        return [
            params_to_unit_scale(processor_params, processor_name)
            for processor_params, processor_name in zip(params, processor_name)
        ]

    processor = processor_by_name[processor_name]
    return {
        param.name: param.to_unit_scale(params[param.name])
        for param in processor.PARAMS
    }


def params_from_unit_scale(params, processor_name):
    if not params or not processor_name:
        return None

    if isinstance(processor_name, list):
        return [
            params_from_unit_scale(processor_params, processor_name)
            for processor_params, processor_name in zip(params, processor_name)
        ]

    processor = processor_by_name[processor_name]
    return {
        param.name: param.from_unit_scale(params[param.name])
        for param in processor.PARAMS
    }


def processor_names_from_graph_config(graph_config):
    if not graph_config:
        return None

    return [
        processor_names_from_graph_config(processor_config)
        if isinstance(processor_config, list)
        else processor_config["name"]
        for processor_config in graph_config
    ]


def processor_names_to_graph_config(processor_names):
    if not processor_names:
        return None

    return [
        processor_names_to_graph_config(processor_name)
        if isinstance(processor_name, list)
        else {"name": processor_name}
        for processor_name in processor_names
    ]


def processors_to_graph_config(processors):
    return [
        processors_to_graph_config(processor)
        if isinstance(processor, list)
        else {"name": processor.NAME}
        for processor in processors
    ]


def get_graph_params(graph_config):
    if not graph_config:
        return None

    return [
        get_graph_params(processor_config)
        if isinstance(processor_config, list)
        else (
            processor_config["params"]
            if "params" in processor_config
            else default_param_values(processor_by_name[processor_config["name"]])
        )
        for processor_config in graph_config
    ]


def init_graph_state(graph_config):
    if not graph_config:
        return None

    return [
        init_graph_state(processor_config)
        if isinstance(processor_config, list)
        else processor_by_name[processor_config["name"]].init_state()
        for processor_config in graph_config
    ]


# Returns (params, state)
def graph_config_to_carry(graph_config):
    if not graph_config:
        return None, None

    return get_graph_params(graph_config), init_graph_state(graph_config)


def set_state_recursive(state, key, value):
    if not state:
        return

    if isinstance(state, list):
        for state_item in state:
            set_state_recursive(state_item, key, value)
    else:
        state[key] = value



import jax.numpy as jnp
from jax import jit, lax
from jax.ops import index_update

from jaxdsp.param import Param

NAME = "Allpass Filter"
PARAMS = [Param("feedback", 0.0)]
PRESETS = {}


def init_state(buffer_size=20):
    return {
        "buffer": jnp.zeros(buffer_size),
        "buffer_index": 0,
        "filter_store": 0.0,
    }


@jit
def tick(carry, x):
    params, state = carry
    buffer_out = state["buffer"][state["buffer_index"]]
    state["buffer"] = index_update(
        state["buffer"], state["buffer_index"], x + buffer_out * params["feedback"]
    )
    state["buffer_index"] += 1
    state["buffer_index"] %= state["buffer"].size
    out = -x + buffer_out
    return carry, out


@jit
def tick_buffer(carry, X):
    return lax.scan(tick, carry, X)



import jax.numpy as jnp


class Param:
    def __init__(
        self, name, default_value=0.0, min_value=0.0, max_value=1.0, log_scale=False
    ):
        self.name = name
        self.default_value = default_value
        self.min_value = min_value
        self.max_value = max_value
        self.log_scale = log_scale

    def from_unit_scale(self, unit_scale_value):
        return self.min_value + unit_scale_value * (self.max_value - self.min_value)

    def to_unit_scale(self, value):
        return (value - self.min_value) / (self.max_value - self.min_value)

    def serialize(self):
        serialized = self.__dict__
        if isinstance(self.default_value, jnp.DeviceArray):
            serialized["default_value"] = self.default_value.tolist()
        return serialized



import numpy as np
from jax import value_and_grad, jit
from jax.tree_util import tree_map, tree_multimap

from jaxdsp import processor_graph
from jaxdsp.loss import LossOptions, loss_fn
from jaxdsp.optimizers import create_optimizer
from jaxdsp.processors import (
    processor_names_from_graph_config,
    processor_by_name,
    graph_config_to_carry,
    params_to_unit_scale,
    params_from_unit_scale,
    get_graph_params,
    processor_names_to_graph_config,
)


@jit
def mean_loss_and_grads(loss, grads):
    return np.mean(loss), tree_map(lambda grad: np.mean(grad, axis=0), grads)


class LossHistoryAccumulator:
    def __init__(self, params_init):
        self.params_history = tree_map(lambda param: [param], params_init)
        self.loss_history = []

    def after_step(self, loss, new_params):
        self.loss_history.append(loss)
        self.params_history = tree_multimap(
            lambda new_param, params: params + [new_param],
            new_params,
            self.params_history,
        )


def float_params(params):
    return tree_map(
        lambda param: param
        if hasattr(param, "ndim") and param.ndim > 0
        else float(param),
        params,
    )


class IterativeTrainer:
    def __init__(
        self,
        graph_config=None,
        loss_options=None,
        optimizer_options=None,
        track_history=False,
    ):
        self.step_num = 0
        self.loss = 0.0
        self.track_history = track_history
        self.params = None
        self.processor_names = None
        self.set_optimizer_options(optimizer_options)
        self.set_graph_config(graph_config)
        self.set_loss_options(loss_options)

    def set_graph_config(self, graph_config):
        self.processor_names = processor_names_from_graph_config(graph_config)
        self.set_carry(graph_config_to_carry(graph_config))

    def set_carry(self, carry):
        if carry:
            params, self.state = carry
            self.params = params or get_graph_params(
                processor_names_to_graph_config(self.processor_names)
            )
            self.step_evaluator = LossHistoryAccumulator(self.params)
        else:
            self.params, self.state = None, None
            self.step_evaluator = None

        self.update_opt_state()

    def set_optimizer_options(self, optimizer_options):
        self.optimizer = (
            create_optimizer(
                optimizer_options.get("name"), optimizer_options.get("params")
            )
            if optimizer_options
            else create_optimizer()
        )

        if self.processor_names:
            self.state = [
                processor_by_name[processor_name].init_state()
                for processor_name in self.processor_names
            ]

        self.update_opt_state()

    def update_opt_state(self):
        self.opt_state = self.optimizer.init(
            params_to_unit_scale(self.params, self.processor_names)
        )

    def set_loss_options(self, loss_options):
        self.loss_options = loss_options or LossOptions()

        def processor_loss(unit_scale_params, state, X, Y_target):
            params = params_from_unit_scale(unit_scale_params, self.processor_names)
            carry, Y_estimated = processor_graph.tick_buffer(
                (params, state), X, self.processor_names
            )
            if Y_estimated.shape == Y_target.shape[::-1]:
                Y_estimated = Y_estimated.T  # TODO should eventually remove this check
            return (
                loss_fn(Y_estimated, Y_target, self.loss_options),
                carry[1],  # return state as aux
            )

        self.grad_fn = jit(value_and_grad(processor_loss, has_aux=True))

    def step(self, X, Y_target):
        if not self.processor_names:
            return

        params_unit = params_to_unit_scale(self.params, self.processor_names)
        (self.loss, self.state), self.grads = self.grad_fn(
            params_unit,
            self.state,
            X,
            Y_target,
        )
        self.opt_state = self.optimizer.update(
            self.step_num, self.grads, self.opt_state
        )
        self.step_num += 1
        self.params = params_from_unit_scale(
            self.optimizer.get_params(self.opt_state), self.processor_names
        )
        if self.step_evaluator:
            self.step_evaluator.after_step(self.loss, self.params)

    def float_params(self):
        return float_params(self.params)

    def params_and_loss(self):
        return {
            "params": self.float_params(),
            "loss": float(self.loss),
        }


def evaluate(carry_estimated, carry_target, processor, X):
    carry_estimated, Y_estimated = processor.tick_buffer(carry_estimated, X)
    carry_target, Y_target = processor.tick_buffer(carry_target, X)
    return Y_estimated, Y_target



import jax.numpy as jnp
from jax import jit

from jaxdsp.processors.constants import DEFAULT_SAMPLE_RATE
from jaxdsp.param import Param

NAME = "Sine Wave"
PARAMS = [
    Param("frequency_hz", 440.0, 55.0, 7_040, log_scale=True),
]
PRESETS = {}


def init_state():
    return {"phase_radians": 0.0, "sample_rate": DEFAULT_SAMPLE_RATE}


@jit
def tick(carry, x):
    raise "single-sample tick method not implemented for sine_wave"


@jit
def tick_buffer(carry, X):
    params, state = carry
    # Add an extra sample to determine phase for start of next buffer.
    t = jnp.arange(X.size + 1) / state["sample_rate"]
    x = params["frequency_hz"] * 2 * jnp.pi * t + state["phase_radians"]
    state["phase_radians"] = x[-1] % (2 * jnp.pi)
    return carry, jnp.sin(x[:-1])



from collections.abc import Iterable
from matplotlib import pyplot as plt
import numpy as np

from jaxdsp import training


def plot_filter(fig, X, Y, Y_estimated, Y_reference, title):
    column_titles = (
        ["Target", "Reference Implementation", "Estimated"]
        if Y_reference is not None
        else ["Target", "Estimated"]
    )
    row_titles = ["Input", "Output"]
    Ys = [Y, Y_reference, Y_estimated] if Y_reference is not None else [Y, Y_estimated]
    axes = fig.subplots(2, len(column_titles), sharex=True)
    fig.suptitle(title, size=16)
    for ax, column_title in zip(axes[0], column_titles):
        ax.set_title(column_title)
    for ax, row_title in zip(axes[:, 0], row_titles):
        ax.set_ylabel(row_title, size="large")
    for i, ax_column in enumerate(axes.T):
        in_plot = ax_column[0]
        # in_plot.stem(X, basefmt=' ')
        in_plot.plot(X)
        in_plot.set_ylim([X.min() - 0.1, X.max() + 0.1])

        out_plot = ax_column[1]
        # out_plot.stem(Ys[i], basefmt=' ')
        out_plot.plot(Ys[i])
        out_plot.set_ylim([Ys[i].min() - 0.1, Ys[i].max() + 0.1])


def plot_loss(fig, loss_history):
    plot = fig.subplots(1, 1)
    plot.plot(loss_history)
    plot.set_xlabel("Batch")
    plot.set_ylabel("Loss")
    plot.set_title("Loss over time", size=16)
    plot.set_yscale("log", base=10)
    plot.autoscale(tight=True)


def plot_processor(fig, processor_name, params_target, params_history):
    param_groups = []  # list of lists of (key, label, scalar_param_value)
    # All non-list params go into the first (and potentially only) param group (column).
    single_params = [
        (key, key, param)
        for (key, param) in params_target.items()
        if not isinstance(param, Iterable)
    ]
    if len(single_params) > 0:
        param_groups.append(single_params)
    # Each list param gets its own param group (column).
    for (key, params) in [
        (key, params)
        for (key, params) in params_target.items()
        if isinstance(params, Iterable)
    ]:
        param_groups.append(
            [(key, "${}_{}$".format(key, i), param) for i, param in enumerate(params)]
        )
    num_rows, num_cols = max([len(param_group) for param_group in param_groups]), len(
        param_groups
    )
    axes = fig.subplots(num_rows, num_cols, sharex=True)
    if num_rows == 1 and num_cols == 1:
        axes = np.expand_dims(axes, axis=0)
    if len(axes.shape) == 1:
        axes = np.expand_dims(axes, axis=1)
    fig.suptitle(processor_name, fontsize="x-large")
    for param_group_i, param_group in enumerate(param_groups):
        for param_i, (key, label, param) in enumerate(param_group):
            plot = axes[param_i][param_group_i]
            plot.set_title(label)
            plot.axhline(y=param, c="g", linestyle="--", label="Actual params")
            param_history = params_history[key]
            if isinstance(param_history[0], Iterable):
                param_history = np.asarray(param_history)[:, param_i]
            plot.plot(param_history, c="b", label="Estimated params")
            if param_group_i == 0:
                plot.set_ylabel("Value")
            if param_i == 0:
                plot.legend()
            if param_i == len(param_group) - 1:
                plot.set_xlabel("Batch")
            plot.autoscale()


def plot_processors(
    fig, params_targets, params_histories, processor_names, is_row=True
):
    subfigs = fig.subfigures(
        1 if is_row else len(processor_names),
        len(processor_names) if is_row else 1,
        wspace=0.07,
    )

    if len(processor_names) == 1:
        subfigs = [subfigs]

    for subfig, processor_name, params_target, params_history in zip(
        subfigs, processor_names, params_targets, params_histories
    ):
        if isinstance(processor_name, list):
            # Nested processors are processoed in parallel, so display in nested column
            plot_processors(
                subfig, params_target, params_history, processor_name, not is_row
            )
        else:
            plot_processor(
                subfig,
                processor_name,
                params_target,
                params_history,
            )


def plot_train(
    trainer,
    params,
    X,
    Y,
    Y_estimated,
    Y_reference,
    title=None,
    plot_loss_history=True,
    plot_params_history=True,
):
    num_rows = 1
    if plot_loss_history:
        num_rows += 1
    if plot_params_history:
        num_rows += 1

    fig = plt.figure(constrained_layout=True, figsize=(16, num_rows * 4))
    height_ratios = None
    if plot_loss_history:
        height_ratios = [0.75, 0.5, 1] if num_rows == 3 else [0.25, 1]
    subfigs = fig.subfigures(num_rows, 1, wspace=0.07, height_ratios=height_ratios)
    subfig_i = 0

    plot_filter(subfigs[subfig_i], X, Y, Y_estimated, Y_reference, title)
    subfig_i += 1

    if plot_loss_history:
        plot_loss(subfigs[subfig_i], trainer.step_evaluator.loss_history)
        subfig_i += 1

    if plot_params_history:
        plot_processors(
            subfigs[subfig_i],
            training.float_params(params),
            trainer.step_evaluator.params_history,
            trainer.processor_names,
        )
        subfig_i += 1


def plot_optimization(
    processor_config, Xs, params_inits, params_target, varying_param_name, steps=50
):
    """Used to investigate the shape of the gradient curve for a single target value across
    a range of different initial values.
    `params_inits` should be a list of dicts - one for each parameter tree to initialize with
    `params_target` should be a single target params dict
    """

    params_estimated = {key: [] for key in params_target.keys()}
    initial_losses = []
    losses = []
    x = [
        param_init[varying_param_name] for param_init in params_inits
    ]  # TODO only pass in one to vary
    for params_init in params_inits:
        trainer = training.IterativeTrainer(
            processor_config,
            training.Config(),
            params_init,
            track_history=True,
        )
        for i in range(steps):
            X = Xs[i % Xs.shape[0]]
            carry_target, Y_target = trainer.processor.tick_buffer(carry_target, X)
            trainer.step(X, Y_target)
        for key, param_estimated in params_estimated_i[processor.NAME].items():
            params_estimated[key].append(param_estimated)
        initial_losses.append(loss_history[0])
        losses.append(loss_history[-1])
    _, axes = plt.subplots(len(params_target) + 2, 1, figsize=(14, 12))
    for param_i, (label, param_target) in enumerate(params_target.items()):
        estimated_params_plot = axes[param_i]
        estimated_params_plot.plot(x, params_estimated[label], linewidth=3)
        estimated_params_plot.set_title("Estimated {}".format(label), size=18)
        estimated_params_plot.set_ylabel(
            "Estimated after {} steps".format(steps), size=11
        )
        estimated_params_plot.axhline(
            param_target, linestyle="--", c="r", label="Target index"
        )
        estimated_params_plot.legend()
    initial_loss_plot = axes[-2]
    initial_loss_plot.plot(x, initial_losses, linewidth=3)
    initial_loss_plot.set_title("Initial loss", size=18)
    initial_loss_plot.set_ylabel("Initial MSE loss", size=11)
    optimized_loss_plot = axes[-1]
    optimized_loss_plot.plot(x, losses, linewidth=3)
    optimized_loss_plot.set_title("Optimized loss", size=18)
    optimized_loss_plot.set_xlabel("Initial guess for {} value".format(label), size=16)
    optimized_loss_plot.set_ylabel("MSE loss after {} steps".format(steps), size=11)
    for axis in axes:
        axis.grid(True)
    plt.tight_layout()



import jax.numpy as jnp
from jax import jit

from jaxdsp.param import Param

NAME = "Clip"
PARAMS = [Param("min", -1.0, -1.0, 1.0), Param("max", 1.0, -1.0, 1.0)]
PRESETS = {}


def init_state():
    return {}


@jit
def tick(carry, x):
    params, _ = carry
    return carry, jnp.clip(x, params["min"], params["max"])


@jit
def tick_buffer(carry, X):
    return tick(carry, X)



import jax.numpy as jnp

from jaxdsp.processors import processor_by_name


# This module does not actually support full graph connectivity.
# For simplicity, only series & parallel processing is supported.
# A wide variety of common DSP techniques can be implemented by nesting series and parallel processor groups.
# Keeping this restriction (at least for now) means the graph structure can be represented and serialized
# simply as a list, each element of which is a processor or a list.
# This also simplifies the UI since the "graph" connectivity is implicit in the positional order.
# Thus, we don't need to support arbitrary connectivity, and can instead use a simple drag-and-drop UX paradigm.


def tick_buffer_series(carry, X, processor_names):
    params, state = carry

    Y = X
    for i, processor_name in enumerate(processor_names):
        processor_state = state[i]
        processor_params = params[i]
        processor_carry, Y = processor_by_name[processor_name].tick_buffer(
            (processor_params, processor_state), Y
        )
        state[i] = processor_carry[1]

    return carry, Y


def tick_buffer_parallel(carry, X, processor_names):
    params, state = carry

    Y = jnp.zeros(X.shape)
    for i, processor_name in enumerate(processor_names):
        processor_state = state[i]
        processor_params = params[i]
        processor_carry, Y_i = processor_by_name[processor_name].tick_buffer(
            (processor_params, processor_state), X
        )
        Y += Y_i
        state[i] = processor_carry[1]

    return carry, Y


# `processor_names`, and the params/state tuples in `carry`, are each lists,
# each element of which can be a processor or a list.
# The top-level list is interpreted as a series-connected chain,
# and every nested list is interpreted as a parallel-connected chain.
# E.g. `[["Sine Wave", "Sine Wave"], "Allpass Filter"]`
# is two parallel sine wave processors followed by an allpass filter.
def tick_buffer(carry, X, processor_names):
    params, state = carry

    Y = X
    for i, processor_name in enumerate(processor_names):
        processor_state = state[i]
        processor_params = params[i]
        processor_carry, Y = (
            tick_buffer_parallel((processor_params, processor_state), Y, processor_name)
            if isinstance(processor_name, list)
            else processor_by_name[processor_name].tick_buffer(
                (processor_params, processor_state), Y
            )
        )
        state[i] = processor_carry[1]

    return carry, Y



import jax.numpy as jnp
from jax.experimental import optimizers
from jax.experimental.optimizers import optimizer
from jax.tree_util import tree_map

from jaxdsp.param import Param

step_size_param = Param("step_size", 0.05, 1e-9, 0.2)

# Clip all params to [0.0, 1.0] (params are all normalized to unit scale before passing to gradient fn).
# TODO should also add a loss component to strongly encourage params into this range
@optimizer
def param_clipping_optimizer(init, update, get_params):
    # Note that these are the params being optimized, NOT the optimizer params :)
    def get_clipped_params(state):
        params = get_params(state)
        return tree_map(lambda param: jnp.clip(param, 0.0, 1.0), params)

    return init, update, get_clipped_params


class Optimizer:
    def __init__(self, definition, param_values=None):
        self.definition = definition
        self.set_param_values(param_values)

    def set_param_values(self, param_values=None):
        self.param_values = {
            param.name: (param_values or {}).get(param.name) or param.default_value
            for param in self.definition.PARAMS
        }
        init, update, get_params = self.definition.FUNCTION(
            *[self.param_values[param.name] for param in self.definition.PARAMS]
        )
        self.init, self.update, self.get_params = param_clipping_optimizer(
            init, update, get_params
        )

    def serialize(self):
        return {
            "name": self.definition.NAME,
            "param_definitions": [
                param.serialize() for param in self.definition.PARAMS
            ],
            "params": self.param_values,
        }


class AdaGrad:
    NAME = "AdaGrad"
    PARAMS = [step_size_param, Param("momentum", 0.9)]
    FUNCTION = optimizers.adagrad


class Adam:
    NAME = "Adam"
    PARAMS = [step_size_param, Param("b1", 0.9), Param("b2", 0.999)]
    FUNCTION = optimizers.adam


class Adamax:
    NAME = "Adamax"
    PARAMS = [step_size_param, Param("b1", 0.9), Param("b2", 0.999)]
    FUNCTION = optimizers.adamax


class Nesterov:
    NAME = "Nesterov"
    PARAMS = [step_size_param, Param("mass", 0.5)]
    FUNCTION = optimizers.nesterov


class RmsProp:
    NAME = "RMSProp"
    PARAMS = [step_size_param, Param("gamma", 0.9)]
    FUNCTION = optimizers.rmsprop


class Sgd:
    NAME = "SGD"
    PARAMS = [step_size_param]
    FUNCTION = optimizers.sgd


all_optimizer_definitions = [AdaGrad, Adam, Adamax, Nesterov, RmsProp, Sgd]


def create_optimizer(name=None, param_values=None):
    definition = next(
        (
            definition
            for definition in all_optimizer_definitions
            if definition.NAME == name
        ),
        RmsProp,
    )

    return Optimizer(definition, param_values)



# Fractional delay line using linear interpolation

# The `delay_samples` param cannot be optimized outside of the +/- 1 sample range between
# init and target.
# This is because the param maps directly to an `index_update` operation,
# which requires converting the parameter to an integer, which is a non-differentiable operation.
# Linear interpolation helps by allowing successful optimization within the +/- 1 sample
# range, but I have not found a solution that allows the gradient signal to propagate
# across the entire indexing range.
# (See "Differentiable array indexing" notebook for more details.)

import math
import jax.numpy as jnp
import numpy as np
from jax import jit, lax
from jax.ops import index, index_update

from jaxdsp.processors.constants import DEFAULT_SAMPLE_RATE
from jaxdsp.param import Param


# Stay inside the range of the the lowest frequency resolved by the fft in the spectral optimizers
MAX_DELAY_LENGTH_MS = 40.0
MAX_DELAY_LENGTH_SAMPLES = math.ceil(
    DEFAULT_SAMPLE_RATE * (MAX_DELAY_LENGTH_MS / 1000.0)
)

NAME = "Delay Line"
PARAMS = [
    Param("wet", 1.0),
    Param("delay_ms", MAX_DELAY_LENGTH_MS / 2, 0.0, MAX_DELAY_LENGTH_MS),
]
PRESETS = {}


def init_state():
    return {
        "delay_line": jnp.zeros(MAX_DELAY_LENGTH_SAMPLES),
        "read_sample": 0.0,
        "write_sample": 0.0,
        "sample_rate": DEFAULT_SAMPLE_RATE,
    }


@jit
def tick(carry, x):
    params, state = carry
    write_sample = state["write_sample"]
    read_sample = state["read_sample"]

    state["delay_line"] = index_update(
        state["delay_line"], index[write_sample].astype("int32"), x
    )

    read_sample_floor = read_sample.astype("int32")
    interp = read_sample - read_sample_floor
    y = interp * state["delay_line"][read_sample_floor]
    y += (1 - interp) * state["delay_line"][
        (read_sample_floor + 1) % MAX_DELAY_LENGTH_SAMPLES
    ]

    state["write_sample"] += 1
    state["write_sample"] %= MAX_DELAY_LENGTH_SAMPLES
    state["read_sample"] += 1
    state["read_sample"] %= MAX_DELAY_LENGTH_SAMPLES

    out = x * (1 - params["wet"]) + y * params["wet"]
    return carry, out


@jit
def tick_buffer(carry, X):
    params, state = carry
    delay_samples = jnp.ceil(state["sample_rate"] * (params["delay_ms"] / 1000.0))
    state["read_sample"] = (
        state["write_sample"] - jnp.clip(delay_samples, 0, MAX_DELAY_LENGTH_SAMPLES)
    ) % MAX_DELAY_LENGTH_SAMPLES
    return lax.scan(tick, carry, X)



import jax.numpy as jnp
from jax import jit, lax

from jaxdsp.processors import lowpass_feedback_comb_filter as comb
from jaxdsp.processors import allpass_filter as allpass
from jaxdsp.param import Param

NAME = "Freeverb"
PARAMS = [
    Param("wet", 0.0),
    Param("dry", 1.0),
    Param("width", 0.0),
    Param("damp", 0.0),
    Param("room_size", 0.0, 0.0, 1.2),
]
PRESETS = {
    "flat_space": {
        "wet": 0.3,
        "dry": 0.6,
        "width": 0.5,
        "damp": 0.3,
        "room_size": 1.055,
    },
    "expanding_space": {
        "wet": 0.33,
        "dry": 0.0,
        "width": 0.5,
        "damp": 0.1,
        "room_size": 1.078,
    },
}

fixed_gain = 0.015
scale_wet = 3.0
scale_dry = 2.0
scale_damp = 0.4
scale_room = 0.28
offset_room = 0.7

comb_tunings_l = [1116, 1188, 1277, 1356, 1422, 1491, 1557, 1617]
allpass_tunings_l = [556, 441, 341, 225]
stereo_spread = 23

# comb_tunings_l = [1, 2, 3, 4, 5, 6, 7, 9]
# allpass_tunings_l = [2, 3, 4, 5]


def init_comb_carry(buffer_size):
    return ({"feedback": 0.5, "damp": 0.0}, comb.init_state(buffer_size=buffer_size))


def init_allpass_carry(buffer_size):
    return ({"feedback": 0.5}, allpass.init_state(buffer_size=buffer_size))


def init_state():
    return {
        "combs_l": [init_comb_carry(buffer_size) for buffer_size in comb_tunings_l],
        "combs_r": [
            init_comb_carry(buffer_size + stereo_spread)
            for buffer_size in comb_tunings_l
        ],
        "allpasses_l": [
            init_allpass_carry(buffer_size) for buffer_size in allpass_tunings_l
        ],
        "allpasses_r": [
            init_allpass_carry(buffer_size + stereo_spread)
            for buffer_size in allpass_tunings_l
        ],
    }


@jit
def tick(carry, x):
    params, state = carry

    x_r, x_l = jnp.broadcast_to(x, (2,))  # handle stereo or mono in
    x_combined = (x_l + x_r) * fixed_gain

    out_l = 0.0
    out_r = 0.0

    # comb filters in parallel
    for i in range(len(comb_tunings_l)):
        state["combs_l"][i], out_comb_l = comb.tick(state["combs_l"][i], x_combined)
        state["combs_r"][i], out_comb_r = comb.tick(state["combs_r"][i], x_combined)
        out_l += out_comb_l
        out_r += out_comb_r
    # allpasses in series
    for i in range(len(allpass_tunings_l)):
        state["allpasses_l"][i], out_l = allpass.tick(state["allpasses_l"][i], out_l)
        state["allpasses_r"][i], out_r = allpass.tick(state["allpasses_r"][i], out_r)

    wet = params["wet"] * scale_wet
    dry = params["dry"] * scale_dry
    wet_1 = wet * (params["width"] / 2 + 0.5)
    wet_2 = wet * ((1 - params["width"]) / 2)

    output_l = out_l * wet_1 + out_r * wet_2 + x_l * dry
    output_r = out_r * wet_1 + out_l * wet_2 + x_r * dry
    return carry, jnp.array([output_l, output_r])


@jit
def tick_buffer(carry, X):
    params, state = carry

    room_size = (params["room_size"] * scale_room) + offset_room
    damp = params["damp"] * scale_damp
    for comb_l in state["combs_l"]:
        comb_l[0]["feedback"] = room_size
        comb_l[0]["damp"] = damp
    for comb_r in state["combs_r"]:
        comb_r[0]["feedback"] = room_size
        comb_r[0]["damp"] = damp

    return lax.scan(tick, carry, X)



import jax.numpy as jnp
from jax import jit, lax

from jaxdsp.param import Param

NAME = "FIR Filter"
# TODO how to handle array params in UI?
PARAMS = [Param("B", jnp.concatenate([jnp.array([1.0]), jnp.zeros(4)]))]
PRESETS = {}


def init_state(length=4):
    return {"inputs": jnp.zeros(length)}


@jit
def tick(carry, x):
    params, state = carry
    B = params["B"]
    state["inputs"] = jnp.concatenate([jnp.array([x]), state["inputs"][0:-1]])
    y = B @ state["inputs"]
    return carry, y


@jit
def tick_buffer(carry, X):
    params, _ = carry
    B = params["B"]
    return carry, jnp.convolve(X, B)[: -(B.size - 1)]
    # Impossibly, the following seems to perform about the exact same or even faster for large N?! O_O
    # return lax.scan(tick, carry, X)[1]



import jax.numpy as jnp
from jax import jit, lax

from jaxdsp.param import Param

NAME = "IIR Filter"
# TODO how to handle array params in UI?
PARAMS = [
    Param("B", jnp.concatenate([jnp.array([1.0]), jnp.zeros(4)])),
    Param("A", jnp.concatenate([jnp.array([1.0]), jnp.zeros(4)])),
]
PRESETS = {}


def init_state(length=5):
    return {
        "inputs": jnp.zeros(length),
        "outputs": jnp.zeros(length - 1),
    }


@jit
def tick(carry, x):
    params, state = carry
    B = params["B"]
    A = params["A"]
    state["inputs"] = jnp.concatenate([jnp.array([x]), state["inputs"][0:-1]])
    y = B @ state["inputs"]
    if state["outputs"].size > 0:
        y -= A[1:] @ state["outputs"]
        # Don't optimize the output gain, since it's not commonly used and constraining it to 1 helps training
        # Note that this makes the implementation technically incorrect. We can always uncomment if we want it back.
        # y /= A[0]
        state["outputs"] = jnp.concatenate([jnp.array([y]), state["outputs"][0:-1]])
    return carry, y


@jit
def tick_buffer(carry, X):
    return lax.scan(tick, carry, X)
